# -*- coding: utf-8 -*-
from django.db import models


class StoredEventRecord(models.Model):

    id = models.BigAutoField(primary_key=True)

    # Application name.
    application_name = models.CharField(max_length=32)

    # Originator ID (e.g. an entity or aggregate ID).
    originator_id = models.UUIDField()

    # Position (index) of item in sequence.
    originator_version = models.BigIntegerField()

    # Topic of the item (e.g. path to domain event class).
    topic = models.TextField()

    # State of the item (serialized dict, possibly encrypted).
    state = models.BinaryField()

    class Meta:
        unique_together = (
            ("application_name", "originator_id", "originator_version"),
            ("application_name", "id"),
        )
        db_table = "stored_events"


class SnapshotRecord(models.Model):

    uid = models.BigAutoField(primary_key=True)

    # Application name.
    application_name = models.CharField(max_length=32)

    # Originator ID (e.g. an entity or aggregate ID).
    originator_id = models.UUIDField()

    # Position (index) of item in sequence.
    originator_version = models.BigIntegerField()

    # Topic of the item (e.g. path to domain event class).
    topic = models.TextField()

    # State of the item (serialized dict, possibly encrypted).
    state = models.BinaryField()

    class Meta:
        unique_together = (("application_name", "originator_id", "originator_version"),)
        db_table = "snapshots"


class NotificationTrackingRecord(models.Model):

    uid = models.BigAutoField(primary_key=True)

    # Application name.
    application_name = models.CharField(max_length=32)

    # Upstream application name.
    upstream_application_name = models.CharField(max_length=32)

    # Notification ID.
    notification_id = models.BigIntegerField()

    class Meta:
        unique_together = (
            (
                "application_name",
                "upstream_application_name",
                "notification_id",
            ),
        )
        db_table = "notification_tracking"



# -*- coding: utf-8 -*-
from __future__ import annotations

from contextlib import contextmanager
from functools import wraps
from threading import Lock
from typing import TYPE_CHECKING
from uuid import UUID

import django.db
from django.db import models, transaction
from django.db.backends.signals import connection_created
from django.db.transaction import get_connection
from eventsourcing.persistence import (
    AggregateRecorder,
    ApplicationRecorder,
    DatabaseError,
    DataError,
    IntegrityError,
    InterfaceError,
    InternalError,
    Notification,
    NotSupportedError,
    OperationalError,
    PersistenceError,
    ProcessRecorder,
    ProgrammingError,
    StoredEvent,
    Tracking,
)

from eventsourcing_django.models import NotificationTrackingRecord

if TYPE_CHECKING:
    from typing import Any, Dict, Iterator, List, Optional, Type

    from django.db import ConnectionProxy

journal_modes: Dict[str, str] = {}


def detect_sqlite(connection: ConnectionProxy) -> bool:
    return connection.vendor == "sqlite"


def detect_sqlite_memory_mode(connection: ConnectionProxy) -> bool:
    db_name = str(connection.settings_dict["NAME"])
    return ":memory:" in db_name or "mode=memory" in db_name


def set_journal_mode_wal_on_sqlite_file_db(
    sender: Any, connection: ConnectionProxy, **kwargs: Any
) -> None:
    """Enable integrity constraint with sqlite."""
    if detect_sqlite(connection) and not detect_sqlite_memory_mode(connection):
        db_name = str(connection.settings_dict["NAME"])
        if journal_modes.get(db_name) != "WAL":
            cursor = connection.cursor()
            cursor.execute("PRAGMA journal_mode;")
            mode = cursor.fetchall()
            if mode[0][0].upper() != "WAL":
                cursor.execute("PRAGMA journal_mode=WAL;")
            journal_modes[db_name] = "WAL"


connection_created.connect(set_journal_mode_wal_on_sqlite_file_db)


def errors(f: Any) -> Any:
    @wraps(f)
    def _wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            return f(*args, **kwargs)
        except django.db.InterfaceError as e:
            raise InterfaceError from e
        except django.db.DataError as e:
            raise DataError from e
        except django.db.OperationalError as e:
            raise OperationalError from e
        except django.db.IntegrityError as e:
            raise IntegrityError from e
        except django.db.InternalError as e:
            raise InternalError from e
        except django.db.ProgrammingError as e:
            raise ProgrammingError from e
        except django.db.NotSupportedError as e:
            raise NotSupportedError from e
        except django.db.DatabaseError as e:
            raise DatabaseError from e
        except django.db.Error as e:
            raise PersistenceError from e

    return _wrapper


class DjangoAggregateRecorder(AggregateRecorder):
    def __init__(
        self,
        application_name: str,
        model: Type[models.Model],
        using: Optional[str] = None,
    ):
        super().__init__()
        self.application_name = application_name
        self.model = model
        self.using = using
        connection = get_connection(using=self.using)

        self.lock: Optional[Lock]
        if detect_sqlite(connection) and detect_sqlite_memory_mode(connection):
            self.lock = Lock()
        else:
            self.lock = None

    @contextmanager
    def serialize(self) -> Iterator[None]:
        try:
            if self.lock:
                self.lock.acquire()
            yield
        finally:
            if self.lock:
                self.lock.release()

    @errors
    def insert_events(self, stored_events: List[StoredEvent], **kwargs: Any) -> None:
        with self.serialize():
            with transaction.atomic(using=self.using):
                self._lock_table()
                self._insert_events(stored_events, **kwargs)

    def _lock_table(self) -> None:
        connection = get_connection(using=self.using)
        if connection.vendor == "postgresql":
            cursor = connection.cursor()
            db_table = self.model._meta.db_table
            cursor.execute(f"LOCK TABLE {db_table} IN EXCLUSIVE MODE")

    def _insert_events(self, stored_events: List[StoredEvent], **kwargs: Any) -> None:
        records = []
        for stored_event in stored_events:
            record = self.model(
                application_name=self.application_name,
                originator_id=stored_event.originator_id,
                originator_version=stored_event.originator_version,
                topic=stored_event.topic,
                state=stored_event.state,
            )
            records.append(record)
        for record in records:
            record.save(using=self.using)

    @errors
    def select_events(
        self,
        originator_id: UUID,
        gt: Optional[int] = None,
        lte: Optional[int] = None,
        desc: bool = False,
        limit: Optional[int] = None,
    ) -> List[StoredEvent]:
        with self.serialize():
            q = self.model.objects.using(alias=self.using).filter(
                application_name=self.application_name, originator_id=originator_id
            )
            q = q.order_by(("" if not desc else "-") + "originator_version")
            if gt is not None:
                q = q.filter(originator_version__gt=gt)
            if lte is not None:
                q = q.filter(originator_version__lte=lte)
            if limit is not None:
                q = q[0:limit]
            records = list(q)
        return [
            StoredEvent(
                originator_id=r.originator_id,
                originator_version=r.originator_version,
                topic=r.topic,
                state=bytes(r.state) if isinstance(r.state, memoryview) else r.state,
            )
            for r in records
        ]


class DjangoApplicationRecorder(DjangoAggregateRecorder, ApplicationRecorder):
    @errors
    def select_notifications(self, start: int, limit: int) -> List[Notification]:
        with self.serialize():
            q = self.model.objects.using(alias=self.using).filter(
                application_name=self.application_name,
            )
            q = q.order_by("id")
            q = q.filter(id__gte=start)
            q = q[0:limit]
            records = list(q)
        return [
            Notification(
                id=r.id,
                originator_id=r.originator_id,
                originator_version=r.originator_version,
                topic=r.topic,
                state=bytes(r.state) if isinstance(r.state, memoryview) else r.state,
            )
            for r in records
        ]

    @errors
    def max_notification_id(self) -> int:
        with self.serialize():
            q = self.model.objects.using(alias=self.using).filter(
                application_name=self.application_name,
            )
            q = q.order_by("-id")
            try:
                max_id = q[0].id
            except IndexError:
                max_id = 0
        return max_id


class DjangoProcessRecorder(DjangoApplicationRecorder, ProcessRecorder):
    def _insert_events(self, stored_events: List[StoredEvent], **kwargs: Any) -> None:
        super(DjangoProcessRecorder, self)._insert_events(stored_events, **kwargs)
        tracking: Optional[Tracking] = kwargs.get("tracking", None)
        if tracking is not None:
            record = NotificationTrackingRecord(
                application_name=self.application_name,
                upstream_application_name=tracking.application_name,
                notification_id=tracking.notification_id,
            )
            record.save(using=self.using)

    @errors
    def max_tracking_id(self, application_name: str) -> int:
        with self.serialize():
            q = NotificationTrackingRecord.objects.using(alias=self.using).filter(
                application_name=self.application_name,
                upstream_application_name=application_name,
            )
            q = q.order_by("-notification_id")
            try:
                max_id = q[0].notification_id
            except IndexError:
                max_id = 0
        return max_id



# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Mapping

from eventsourcing.persistence import (
    AggregateRecorder,
    ApplicationRecorder,
    InfrastructureFactory,
    ProcessRecorder,
)

from eventsourcing_django.models import SnapshotRecord, StoredEventRecord
from eventsourcing_django.recorders import (
    DjangoAggregateRecorder,
    DjangoApplicationRecorder,
    DjangoProcessRecorder,
)


class Factory(InfrastructureFactory):
    DJANGO_DB_ALIAS = "DJANGO_DB_ALIAS"

    def __init__(self, application_name: str, env: Mapping[str, str]):
        super().__init__(application_name, env)
        self.db_alias = self.getenv(self.DJANGO_DB_ALIAS) or None

    def aggregate_recorder(self, purpose: str = "events") -> AggregateRecorder:
        if purpose == "snapshots":
            model = SnapshotRecord
        else:
            model = StoredEventRecord
        return DjangoAggregateRecorder(
            application_name=self.application_name, model=model, using=self.db_alias
        )

    def application_recorder(self) -> ApplicationRecorder:
        return DjangoApplicationRecorder(
            application_name=self.application_name,
            model=StoredEventRecord,
            using=self.db_alias,
        )

    def process_recorder(self) -> ProcessRecorder:
        return DjangoProcessRecorder(
            application_name=self.application_name,
            model=StoredEventRecord,
            using=self.db_alias,
        )



# -*- coding: utf-8 -*-
# Generated by Django 3.2.5 on 2021-07-31 17:31
from __future__ import annotations

from typing import TYPE_CHECKING

from django.db import migrations, models

if TYPE_CHECKING:
    from typing import List


class Migration(migrations.Migration):

    initial = True

    dependencies: List[str] = []

    operations = [
        migrations.CreateModel(
            name="StoredEventRecord",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                ("application_name", models.CharField(max_length=32)),
                ("originator_id", models.UUIDField()),
                ("originator_version", models.BigIntegerField()),
                ("topic", models.TextField()),
                ("state", models.BinaryField()),
            ],
            options={
                "db_table": "stored_events",
                "unique_together": {
                    ("application_name", "id"),
                    ("application_name", "originator_id", "originator_version"),
                },
            },
        ),
        migrations.CreateModel(
            name="SnapshotRecord",
            fields=[
                ("uid", models.BigAutoField(primary_key=True, serialize=False)),
                ("application_name", models.CharField(max_length=32)),
                ("originator_id", models.UUIDField()),
                ("originator_version", models.BigIntegerField()),
                ("topic", models.TextField()),
                ("state", models.BinaryField()),
            ],
            options={
                "db_table": "snapshots",
                "unique_together": {
                    ("application_name", "originator_id", "originator_version")
                },
            },
        ),
        migrations.CreateModel(
            name="NotificationTrackingRecord",
            fields=[
                ("uid", models.BigAutoField(primary_key=True, serialize=False)),
                ("application_name", models.CharField(max_length=32)),
                ("upstream_application_name", models.CharField(max_length=32)),
                ("notification_id", models.BigIntegerField()),
            ],
            options={
                "db_table": "notification_tracking",
                "unique_together": {
                    ("application_name", "upstream_application_name", "notification_id")
                },
            },
        ),
    ]



# -*- coding: utf-8 -*-
from django.contrib import admin

from .models import NotificationTrackingRecord, SnapshotRecord, StoredEventRecord

admin.site.register(StoredEventRecord)
admin.site.register(SnapshotRecord)
admin.site.register(NotificationTrackingRecord)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['eventsourcing_django', 'eventsourcing_django.migrations']

package_data = \
{'': ['*']}

install_requires = \
['Django>=3.0.0,<4.0.0', 'eventsourcing>=9.1.0,<10.0.0']

setup_kwargs = {
    'name': 'eventsourcing-django',
    'version': '0.1.4',
    'description': 'Python package for eventsourcing with Django.',
    'long_description': '# Event Sourcing with Django\n\nThe Python package `eventsourcing-django` provides a Django app\nthat uses the [Django ORM](https://www.djangoproject.com/)\nto define alternative persistence infrastructure for the\n[Python eventsourcing library](https://github.com/pyeventsourcing/eventsourcing).\n\nThis package is [available on PyPI](https://pypi.org/project/eventsourcing-django/).\n\nThis package is designed and tested to work with version 9.1 of\nthe Python eventsourcing library, Django versions 3.0, 3.1, and\n3.2, and Python versions 3.7, 3.8, 3.9, and 3.10.\n\nThis functionality provided by this package was previously included\nin the Python eventsourcing package, but was moved out to a separate\npackage during development of version 9.\n\nThis is package is maintained by the Python eventsourcing project.\nPlease [raise issues on GitHub](https://github.com/pyeventsourcing/eventsourcing-django/issues)\nand [join the community](https://join.slack.com/t/eventsourcinginpython/shared_invite/enQtMjczNTc2MzcxNDI0LTJjMmJjYTc3ODQ3M2YwOTMwMDJlODJkMjk3ZmE1MGYyZDM4MjIxODZmYmVkZmJkODRhZDg5N2MwZjk1YzU3NmY)\ndiscussion on Slack.\n\n\n## Installation\n\nYou can use `pip` to install the package. It is recommended to install\nPython packages into a Python virtual environment.\n\n    $ pip install eventsourcing-django\n\n\n## Configuration\n\nIf you are using Django 3.0 or 3.1, please add\n`\'eventsourcing_django.apps.EventsourcingConfig\'` to your Django\nproject\'s `INSTALLED_APPS` setting.\n\n    INSTALLED_APPS = [\n        ...\n        \'eventsourcing_django.apps.EventsourcingConfig\',\n    ]\n\n\nIf you are using Django 3.2 or later, you only need to add `\'eventsourcing_django\'`\nto your Django project\'s `INSTALLED_APPS` setting, although the above will work also.\n\n    INSTALLED_APPS = [\n        ...\n        \'eventsourcing_django\',\n    ]\n\n\n## Database migration\n\nTo migrate your database, please run Django\'s `manage.py migrate` command.\n\n    $ python manage.py migrate eventsourcing_django\n\n\n## Event-sourced aggregates and application\n\nYou can develop event-sourced aggregates and applications\nindependently of persistence infrastructure. Please refer\nto the [core library docs](https://eventsourcing.readthedocs.io/)\nfor more information.\n\nThe example below defines an event-sourced aggregate `World`. It\nwill be created with a `history` attribute. The command method\n`make_it_so()` triggers an event `SomethingHappened`\nthat appends the command argument `what` to the `history`.\n\n```python\nfrom eventsourcing.domain import Aggregate, event\n\n\nclass World(Aggregate):\n    def __init__(self):\n        self.history = []\n\n    @event("SomethingHappened")\n    def make_it_so(self, what):\n        self.history.append(what)\n```\n\nThe application class `Universe` has three methods. The method `create_world()`\ncreates a new `World` aggregate. The method `make_it_so()` calls `make_it_so()`\non an existing `World` aggregate. The method `get_world_history()`\nreturns the current `history` value of an existing `World` aggregate.\n\n```python\nfrom eventsourcing.application import Application\n\nclass Universe(Application):\n    def create_world(self):\n        world = World()\n        self.save(world)\n        return world.id\n\n    def make_it_so(self, world_id, what):\n        world = self.repository.get(world_id)\n        world.make_it_so(what)\n        self.save(world)\n\n    def get_world_history(self, world_id):\n        world = self.repository.get(world_id)\n        return world.history\n```\n\n\n## Initialize application object\n\nThe application object brings together the domain model and the\npersistence infrastructure, and provides an interface for views\nand forms.\n\nTo use the Django ORM as the application\'s persistence infrastructure,\nyou must set the application\'s environment variable\n`INFRASTRUCTURE_FACTORY` to `eventsourcing_django.factory:Factory`.\nEnvironment variables can be set in the environment, or set on the\napplication class, or passed in when constructing the application\nobject as seen below.\n\n```python\n# Construct the application.\napp = Universe(env={\n    "INFRASTRUCTURE_FACTORY": "eventsourcing_django.factory:Factory",\n})\n```\n\nYou may wish to construct the application object on a signal\nwhen the Django project is "ready". You can use the `ready()`\nmethod of the `AppConfig` class in the `apps.py` module of a\nDjango app.\n\n\n## Views and forms\n\nAfter migrating the database and constructing the application object,\nthe application object\'s methods can be called. The application object\'s\nmethods may be called from Django views and forms.\n\n```python\n# Call application command methods.\nworld_id = app.create_world()\n\napp.make_it_so(world_id, "dinosaurs")\napp.make_it_so(world_id, "trucks")\napp.make_it_so(world_id, "internet")\napp.make_it_so(world_id, "covid")\n\n# Call application query methods.\nhistory = app.get_world_history(world_id)\nassert history == ["dinosaurs", "trucks", "internet", "covid"]\n```\n\nFor more information, please refer to the Python\n[eventsourcing](https://github.com/johnbywater/eventsourcing) library\nand the [Django](https://www.djangoproject.com/) project.\n',
    'author': 'John Bywater',
    'author_email': 'john.bywater@appropriatesoftware.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://eventsourcing.readthedocs.io/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
